# shield.tcl --

#package provide shield 1.0

set home [string trimright [file dirname [info script]] ./]
set home [file join [pwd] $home]

set web_dir [file dirname $home]
set sh_rules [file join [file dirname $web_dir] rules]
set sh_config [file join [file dirname $web_dir] config]
set sh_tcl $home

unset home
#set sh_rules "/dev/lyris/web/rules"
#set sh_config "/dev/lyris/web/config"
#set sh_tcl "/dev/lyris/web/tcl"
set shield_prefix "shield"
set shield_dir "/${shield_prefix}"

# sets the Mailshield URL httpd direct-URL to the parameter given to it, such as "/shield"
proc Shield_Url {dir} {
    global shield_prefix shield_dir web_dir

    Doc_Root $web_dir/htdocs
    Url_PrefixInstall /shield [list ShieldDomain $shield_prefix]
    #Direct_Url $shield_dir $shield_prefix
    #shield_init
}         

proc shield_init {  } {
    global template_id;
    global sh_tcl
    global user passwd
    set user {}
    set passwd {}

    WriteToLog \n\[[clock format [clock seconds]]\]
    WriteToLog "Initializing MailShield variables..."
    WriteToLog [shield/config/parseallfiles]
#    shield/config/parseallfiles
    shield/config/build/sh_config_many_arrays
    shield/config/build/sh_go_button_cached
    shield/config/build/sh_html_pages
#    open $sh_tcl/template/template.txt
    source $sh_tcl/passcheck.tcl
}

proc shield/config/build/sh_html_pages { } {
    global sh_html_pages sh_tcl sh_html_pages
    global body_bg_color body_text_color 
    global body_link_color body_vlink_color body_alink_color   
    global title_bg_color
    global sh_config_array_filedate sh_config_array_value
    global graphics 
    global template_file
    global web_dir

    # read the config file
    source $sh_tcl/fonts.tcl
    set sh_config_array_filedate(fonts.tcl) [file atime $sh_tcl/fonts.tcl]
    
    set template_file "$web_dir/template/$sh_config_array_value(template)"

    set sh_html_pages(template) [ReadEntireFile $template_file]
    set sh_config_array_filedate(template.txt) [file atime $template_file]
    
    # set the preferences
    regsub "!!MSLOGO!!" $sh_html_pages(template) /images/$ms_logo sh_html_pages(template)
    regsub "!!BODYBGCOLOR!!" $sh_html_pages(template) $body_bg_color sh_html_pages(template)
    regsub "!!BODYTEXTCOLOR!!" $sh_html_pages(template) $body_text_color sh_html_pages(template)
    regsub "!!BODYLINKCOLOR!!" $sh_html_pages(template) $body_link_color sh_html_pages(template)
    regsub "!!BODYVLINKCOLOR!!" $sh_html_pages(template) $body_vlink_color sh_html_pages(template)
    regsub "!!BODYALINKCOLOR!!" $sh_html_pages(template) $body_alink_color sh_html_pages(template)
    regsub "!!TITLEBGCOLOR!!" $sh_html_pages(template) $title_bg_color sh_html_pages(template)
#    set sh_html_pages(templace) [ReadEntireFile "$sh_tcl/template/admin.htm"]
#    set sh_html_pages("admin-menu") [ReadEntireFile "$sh_tcl/template/a-menu.htm"]
}

# Function called for each url. Process the URL and dispatch the corresponding
# tcl function.
# It is a basic copy of TclHttpd function 'DirectDomain' but an authentification
# process has been added.

proc ShieldDomain {prefix sock suffix} {
#    upvar #0 Httpd$sock data
    global sh_tcl shield_prefix
    global Direct
    global template_file

    global sh_config_array_filedate

#    WriteToLog "$prefix $suffix"

    if {$sh_config_array_filedate(template.txt) != [file atime $template_file] || \
         $sh_config_array_filedate(fonts.tcl) != [file atime "$sh_tcl/fonts.tcl"]} {
        shield/config/build/sh_html_pages
        source $sh_tcl/fonts.tcl
    }

    # Initialise direction path for the cache

    set Direct($shield_prefix) /$shield_prefix

    # Perform the authorization process

    set cook [Auth_Check $sock $sh_tcl {}]
    if (![Auth_Verify $sock $cook]) {
        return
    }

    # for graphic version of mailshield
    # return a "404" error if an image was not found
    if {[regexp images $suffix] == 1} {

        catch [set im [open $sh_tcl$suffix]]
        if {[catch [Httpd_ReturnFile $sock image/gif $sh_tcl$suffix] val] == 0} {
            return $val
        } else {
            Httpd_Error $sock 404 "$suffix is not present on this server"
        }
    }

#    if {$suffix == "/"} {
#        set suffix ""
#    }
    DirectDomain $prefix $sock $suffix
}


# MailShield home page
proc shield { } {
    global shield_dir
    append retstring "<html>\n<HEAD>\n"
    append retstring "<META HTTP-EQUIV=\"expires\" CONTENT=\"0\">\n"
    append retstring "</HEAD>\n"
    append retstring "<BODY>\n"
    append retstring "- <a href=\"${shield_dir}/admin\">MailShield Administrators</a><br>\n"
    append retstring "</body></html>\n"
    return $retstring
}

# MailShield admin home page
proc shield/admin { } {
    global sh_config_page_number_total sh_html_pages sh_config_cats sh_config_url_array_no_plus
    set thispage $sh_html_pages(template)
    
    regsub "!!CONTENTS!!" $thispage [shield/config/get_html_toc] thispage
    regsub "!!GO!!" $thispage [shield/config/get_go_button_cached] thispage
 #   regsub "!!TITLECOLOR!!" $thispage "#ff0000" thispage
    regsub "!!FEATURENAME!!" $thispage "MailShield Main Admin Menu" thispage
    regsub "!!NEXT!!" $thispage [shield/config/get_next_url] thispage
    regsub "!!PREVIOUS!!" $thispage [shield/config/get_previous_url] thispage
    regsub "!!PAGETOTAL!!" $thispage $sh_config_page_number_total thispage
    regsub "!!PAGENO!!" $thispage [shield/config/get_page_number] thispage
    regsub "!!TITLE!!" $thispage "MailShield Administrator Main Menu" thispage
    
    foreach cat [lsort [array names sh_config_cats]] {
        append categories "<li><a href=\"$sh_config_url_array_no_plus($cat)\">$cat</a>\n"
    }    
    set body "<h3>Main feature categories:</h3>\n<ul>$categories</ul>\n"
    regsub "!!BODY!!" $thispage $body thispage
    #regsub "!!BODY!!" $thispage $sh_html_pages("admin-menu") thispage

    return $thispage
}

# view a category of settings
proc shield/config/viewcategory { category } {
    global sh_config_page_number_total sh_html_pages sh_config_array_justname sh_config_array_desc sh_config_array_tag 
    global shield_dir sh_config_cats sh_config_url_array
    set thispage $sh_html_pages(template)
    regsub -all -- { } $category "+" cat_href
    
    regsub "!!CONTENTS!!" $thispage [shield/config/get_html_toc $category] thispage
    regsub "!!GO!!" $thispage [shield/config/get_go_button_cached] thispage
#    regsub "!!TITLECOLOR!!" $thispage "#efefef" thispage
    regsub "!!FEATURENAME!!" $thispage "<a href=\"${shield_dir}/admin\">Main Menu</a> : <font size=+2><b>$category</b></font>" thispage
    regsub "!!NEXT!!" $thispage [shield/config/get_next_url $cat_href] thispage
    regsub "!!PREVIOUS!!" $thispage [shield/config/get_previous_url $cat_href] thispage
    regsub "!!PAGETOTAL!!" $thispage $sh_config_page_number_total thispage
    regsub "!!PAGENO!!" $thispage [shield/config/get_page_number $cat_href] thispage
    regsub "!!TITLE!!" $thispage $category thispage
    
    append retstring "<h3>Settings under the category of '$category' :</h3><ul>\n"
    set filenames $sh_config_cats($category)
    foreach filename [lsort $filenames] {
        set justname $sh_config_array_justname($filename)
        set desc $sh_config_array_desc($justname)
        set tag $sh_config_array_tag($justname)
        append retstring "<li> <b><a href=\"$sh_config_url_array($justname)\">$tag</a></b> - $desc<br>\n"
    }
    append retstring "</ul>\n"
    regsub "!!BODY!!" $thispage $retstring thispage

    return $thispage
}

# edit a setting
proc shield/config/edit { setting {highlight {}}} {
    global sh_config_page_number_total sh_html_pages sh_config_array_justname sh_config_array_desc sh_config_array_tag 
    global shield_dir sh_config_cats sh_config_url_array sh_full_text_body sh_config_array_keywords sh_config_array_help
    global sh_config_array_value sh_config_cat_by_setting sh_config_url_array_no_plus sh_file_keywords
    global sh_keyword_url sh_config_array_instructions
    global sh_config sh_config_array_filename sh_config_array_filedate
    set thispage $sh_html_pages(template)
    set tag $sh_config_array_tag($setting)
    
    
    set this_cat $sh_config_cat_by_setting($setting)
    set this_cat_href $sh_config_url_array_no_plus($this_cat)
    set category "<a href=\"$this_cat_href\">$this_cat</a>"
    #set category "cat"  
    
    regsub "!!CONTENTS!!" $thispage [shield/config/get_html_toc $setting] thispage
    regsub "!!GO!!" $thispage [shield/config/get_go_button_cached] thispage
#    regsub "!!TITLECOLOR!!" $thispage "#efefef" thispage
    regsub "!!FEATURENAME!!" $thispage "<a href=\"${shield_dir}/admin\">Main Menu</a> : $category : <font size=+2><b>$tag</b></font>" thispage
    regsub "!!NEXT!!" $thispage [shield/config/get_next_url $setting] thispage
    regsub "!!PREVIOUS!!" $thispage [shield/config/get_previous_url $setting] thispage
    regsub "!!PAGETOTAL!!" $thispage $sh_config_page_number_total thispage
    regsub "!!PAGENO!!" $thispage [shield/config/get_page_number $setting] thispage
    regsub "!!TITLE!!" $thispage $tag thispage

    # Read the config file and keep only the user section in variable lines
    # if the config file has been modified
    if {$sh_config_array_filedate($sh_config_array_filename($setting)) != [file atime $sh_config/$sh_config_array_filename($setting)]} {
        set bool 0
        set lines {}
        foreach line [split [ReadEntireFile $sh_config/$sh_config_array_filename($setting)] \n] {
            if {$bool == 1} {
                append lines $line
                append lines \n
            }
            if {[string first "# Instructions:" $line] == 0} {
                set bool 1
            }
        }

        set lines [string trimright $lines \n]
        set sh_config_array_filedate($sh_config_array_filename($setting) [file atime $sh_config/$sh_config_array_filename($setting)]
        #append retstring "Different dates<br>"
    } else {
        set lines $sh_config_array_value($setting)
        #append retstring "Same dates<br>"
    }
    
    #append retstring "<b>Setting:</b> $tag<br>";
    append retstring "Description: <b>$sh_config_array_desc($setting)</b><br>"
    #append retstring "<b>Keywords:</b> $sh_config_array_keywords($setting)<br>"
    #append retstring "<b>Filename:</b> $setting.txt<p>";
    
    set kwstring "Keywords: "
    foreach kw $sh_file_keywords($setting) {
       append kwstring "<a href=\"$sh_keyword_url($kw)\">$kw</a>, "
    }
    append retstring [string trimright $kwstring " ,"]
    append retstring "<p>"
    
    #append retstring "Instructions: $sh_config_array_instructions($setting)"
    
    #append retstring "<TEXTAREA COLS=40 ROWS=5 NAME=\"okrelayd\">$sh_config_array_value($setting)</TEXTAREA><p>\n"
    #append retstring "<INPUT TYPE=\"submit\" VALUE=\"Save\"><INPUT TYPE=\"reset\" VALUE=\"Reset\">\n"
    #append retstring "<p>"
    
    append retstring "$sh_config_array_help($setting)<br>"

    #highlight the words contained is "highlight" variable
    foreach word $highlight {
        global highlight_color
        regsub -all -nocase ($word) $retstring $highlight_color retstring
    }

    append retstring "<form action=\"/shield/config/save\" method=post>\n\
                <textarea name=\"parameters\" rows=\"6\" cols=\"55\">$lines</textarea><br>\n\
                <input type=\"submit\" value=\"Save\/Modify\" name=>\n\
                <input type=\"hidden\" name=\"setting\" value=\"$setting\"></form>"
    

    #set retstring $sh_full_text_body($setting)
    regsub "!!BODY!!" $thispage $retstring thispage

    return $thispage
}


# full text search
proc shield/admin/search { words } {
    global sh_config_page_number_total sh_html_pages sh_config_url_array sh_full_text_body sh_config_array_tag sh_config_array_desc
    set thispage $sh_html_pages(template) 
    
    # clean up the search words
    regsub -all -- {[^a-zA-Z ]} $words " " words
    
    regsub "!!CONTENTS!!" $thispage [shield/config/get_html_toc] thispage
    regsub "!!GO!!" $thispage [shield/config/get_go_button_cached] thispage
#    regsub "!!TITLECOLOR!!" $thispage "#efefef" thispage
    regsub "!!FEATURENAME!!" $thispage "<font size=+2>Search Results</font>" thispage
    regsub "!!NEXT!!" $thispage $sh_config_url_array() thispage
    regsub "!!PREVIOUS!!" $thispage $sh_config_url_array() thispage
    regsub "!!PAGETOTAL!!" $thispage $sh_config_page_number_total thispage
    regsub "!!PAGENO!!" $thispage 1 thispage
    regsub "!!TITLE!!" $thispage "Search results for: [search_words_describe $words]" thispage

    # perform the search    
    set highest 0
    foreach word $words {
        if {[string length $word] >= 2} {
            foreach page_to_search [array names sh_full_text_body] {
                if { [regexp -nocase $word $sh_full_text_body($page_to_search)] == 1} {
                    set current {}
                    catch { set current $found($page_to_search) }
                    lappend current $word
                    if {[llength $current] > $highest} {
                        set highest [llength $current]
                    }
                    set found($page_to_search) $current
                }
            }
        }
    } 

    # decide on whether the search results will display additional info or not (yes, if more than 1 word was searched for)
    set additional_info {}
    if {[llength $words] > 1} {
        set get_additional 1
    } else {
        set get_additional 0
    }

    # build the results text, in a sorted manner
    foreach hit [array names found] {
        if {$get_additional == 1} {
            set additional_info "<br><font size=-1>[search_words_describe $found($hit)]<br></font>"
        }
        
        set url $sh_config_url_array($hit)
        set search_hits($sh_config_array_tag($hit)) "<a href=\"$url\\&highlight=[join $found($hit) +]\">$sh_config_array_tag($hit)</a> $sh_config_array_tag($hit) - $sh_config_array_desc($hit)$additional_info<p>\n"
        set search_scores($sh_config_array_tag($hit)) [llength $found($hit)]
    }
    
    # put a header on the search results
    append search_results ""
    if {[array size search_hits] > 1} {
        set search_results "<h3>Your search found [array size search_hits] pages for: [search_words_describe $words]</h3>\n\n"
    } elseif {[array size search_hits] == 1} {
        set search_results "<h3>Your search found 1 page for: [search_words_describe $words]</h3>\n\n"
    } else {
        if {[string length $words] > 0 } {
            set search_results "<h3>Your search found no hits for: [search_words_describe $words]</h3>\n"
        } else {
            set search_results "<h3>Your did not specify any words to search for!</h3>\n"
        }
    }

    # actually return the search results now
    
    for {set score $highest} {$score > 0} {incr score -1} {
        set these_search_results {}
        set num_results 0
        foreach search_hit [lsort [array names search_hits]] {
            if {$search_scores($search_hit) == $score} {
                incr num_results
                append these_search_results "$search_hits($search_hit)"
            }
        }
        
        if {$num_results > 0} {
            if {$num_results == 1} { set plural_pages "page" } else { set plural_pages "pages" }
            if {$score == 1} { set plural_score "word" } else { set plural_score "words" }
            append search_results "<hr><b>$num_results $plural_pages found with $score matching $plural_score:</b><p>\n"
            append search_results $these_search_results 
        }
    }
                                             

    regsub "!!BODY!!" $thispage $search_results thispage

    return $thispage
}

proc search_words_describe { words } {
    set word_number 0
    set results {}
    foreach word $words {
        incr word_number
        if {$word_number > 1} {
            append results " and "
        }
        append results "'$word'"
    }
    
    return $results
}

# displays the named file
proc shield/config/showfile { file } {
    global sh_config
    return "<pre>\n[ReadEntireFile "$sh_config/$file"]\n</pre>"
}

proc shield/admin/index { } {
    global sh_config_page_number_total sh_html_pages sh_config_url_array sh_full_text_body sh_config_array_tag sh_config_array_desc
    set thispage $sh_html_pages(template)
    
    regsub "!!CONTENTS!!" $thispage [shield/config/get_html_toc] thispage
    regsub "!!GO!!" $thispage [shield/config/get_go_button_cached] thispage
#    regsub "!!TITLECOLOR!!" $thispage "#efefef" thispage
    regsub "!!FEATURENAME!!" $thispage "<font size=+2>Index</font>" thispage
    regsub "!!NEXT!!" $thispage $sh_config_url_array() thispage
    regsub "!!PREVIOUS!!" $thispage $sh_config_url_array() thispage
    regsub "!!PAGETOTAL!!" $thispage $sh_config_page_number_total thispage
    regsub "!!PAGENO!!" $thispage 1 thispage
    regsub "!!TITLE!!" $thispage "Index" thispage
    
    foreach cat [lsort [array names sh_config_cats]] {
        append categories "<li><a href=\"$sh_config_url_array_no_plus($cat)\">$cat</a>\n"
    }    
    set body [shield/admin/indextext]
    regsub "!!BODY!!" $thispage $body thispage
    return $thispage
}

proc shield/admin/index/word { { word {} } } {
    global sh_config_page_number_total sh_html_pages sh_config_url_array sh_full_text_body 
    global shield_dir sh_config_array_tag sh_config_array_desc
    set thispage $sh_html_pages(template)
    
    regsub "!!CONTENTS!!" $thispage [shield/config/get_html_toc] thispage
    regsub "!!GO!!" $thispage [shield/config/get_go_button_cached] thispage
#    regsub "!!TITLECOLOR!!" $thispage "#efefef" thispage
    regsub "!!FEATURENAME!!" $thispage "<font size=+2><a href=\"${shield_dir}/admin/index\">Index</a> : $word</font>" thispage
    regsub "!!NEXT!!" $thispage $sh_config_url_array() thispage
    regsub "!!PREVIOUS!!" $thispage $sh_config_url_array() thispage
    regsub "!!PAGETOTAL!!" $thispage $sh_config_page_number_total thispage
    regsub "!!PAGENO!!" $thispage 1 thispage
    regsub "!!TITLE!!" $thispage "Index" thispage
    
    foreach cat [lsort [array names sh_config_cats]] {
        append categories "<li><a href=\"$sh_config_url_array_no_plus($cat)\">$cat</a>\n"
    }    
    set body [shield/admin/indextermtext $word]
    regsub "!!BODY!!" $thispage $body thispage
    return $thispage
}

# lists all the files in the "rules" directory
proc shield/admin/indextext { } {
    global sh_config_array_keywords_inverted sh_keyword_url sh_config_url_array sh_config_array_desc
    global sh_config_array_tag
    
    # make an array of lowercase terms so that the sort order is case insensitive
    foreach kw [array names sh_config_array_keywords_inverted] {
        set no_case([string tolower $kw]) $kw
    }
    
    set retstring "<h3>Index Terms</h3><ul>"
    foreach kw2 [lsort [array names no_case]] {
        set kw $no_case($kw2)
        append retstring "<a href=\"$sh_keyword_url($kw)\">$kw</a><br>\n"
    }
    append retstring "</ul>"

    return $retstring
}

# lists all the files in the "rules" directory
proc shield/admin/indextermtext { word } {
    global sh_config_array_keywords_inverted sh_keyword_url sh_config_url_array sh_config_array_desc
    global sh_config_array_tag
    
    if {[string length $word] == 0 } {
        set retstring "<h3>Index Terms</h3><ul>"
        foreach kw [lsort [array names sh_config_array_keywords_inverted]] {
            append retstring "<li><a href=\"$sh_keyword_url($kw)\">$kw</a>\n"
        }
        append retstring "</ul>"
    } else {
        set retstring "<h3>The index term \"$word\" is found on these pages:</h3><ul>"
        foreach file $sh_config_array_keywords_inverted($word) {
            append retstring "<li><a href=\"$sh_config_url_array($file)\"><b>$sh_config_array_tag($file)</b></a>: $sh_config_array_desc($file)\n"
        }
        append retstring "</ul>"
    }

    return $retstring
}


# fills the global variables with the parsed values of all the config files
proc shield/config/parseallfiles { } {
    global sh_config_cats
    global sh_config_array
    global sh_config_files
    global sh_config
    global sh_full_text_body sh_config_array_justname
	# run_as_daemon is passed by the MailShield C++ code to know if MailShield runs in daemon mofr or not
    global run_as_daemon
    
    catch { unset sh_config_cats }
    catch { unset sh_config_array }
    
    set sh_config_files [shield/config/listfiles]
    foreach configfile $sh_config_files {
		# parse the configuration files in the "config" directory
        if {$configfile != "password.txt"} {
            if {[catch {shield/config/parsefile $configfile}] == 0} {
                set sh_full_text_body($sh_config_array_justname($configfile)) [ReadEntireFile "$sh_config/$configfile"]
            } else {
				# if a file does not have the correct format, then print a warning message or write it in
				# the Web interface log file if MailShield is running as daemon mode
                if {$run_as_daemon == 0} {
                    puts stderr "Bypassing config\\$configfile: the file does not have a good format."
                } else {
                    WriteToLog "Bypassing config\\$configfile: the file does not have a good format."
                }
            }
        }
    }
    return "Done";
}

# show all categories as a <pre> block
proc shield/config/showcategories { } {
    global sh_config_array
    global sh_config_cats
    global sh_config_array_filename sh_config_array_desc sh_config_array_cat sh_config_array_form sh_config_array_justname sh_config_array_tag
    
    foreach cat [lsort [array names sh_config_cats]] {
        append retstring "$cat\n"
        set filenames $sh_config_cats($cat)
        foreach filename [lsort $filenames] {
            set justname $sh_config_array_justname($filename)
            set desc $sh_config_array_desc($justname)
            set tag $sh_config_array_tag($justname)
            append retstring "$justname - $tag - $desc\n"
        }
        append retstring "-----\n"
    }
    return "<pre>$retstring</pre>"
}

# returns the table of contents as an HTML document -- current document should be passed as "setting"
proc shield/config/get_html_toc { {setting {}} } {
    global sh_config_array
    global sh_config_cats
    global shield_dir
    global sh_config_array_filename sh_config_array_desc sh_config_array_cat sh_config_array_form sh_config_array_justname sh_config_array_tag
    global sh_config_url_array 
        
    foreach cat [lsort [array names sh_config_cats]] {
        regsub -all -- { } $cat "+" cat_href
        if { $setting == $cat } { 
            append retstring "<p></font><font size=+1><b>$cat</b></font><font face=arial size=-1><br>\n"
        } else {
            #append retstring "<p><b><a href=\"${shield_dir}/config/viewcategory?category=$cat_href\">$cat</a></b><br>\n"
            append retstring "<p><b><a href=\"$sh_config_url_array($cat_href)\">$cat</a></b><br>\n"
        }

        append retstring "<font face=arial size=-1>\n"
        set filenames $sh_config_cats($cat)
        foreach filename [lsort $filenames] {
            set justname $sh_config_array_justname($filename)
            set desc $sh_config_array_desc($justname)
            set tag $sh_config_array_tag($justname)
            if { $setting == $justname } { 
                append retstring "- <b></font><font size=+1>$tag</font><font face=arial size=-1></b><br>\n"
            } else {
                append retstring "- <a href=\"$sh_config_url_array($justname)\">$tag</a><br>\n"
            }
        }
        append retstring "</font>\n"
    }
    return $retstring
}

# determines the next HREF item in the TOC
proc shield/config/get_next_url { { setting {} } } {
    global sh_config_url_array 
    return $sh_config_url_array([shield/config/get_next $setting])
}

# determines the previous HREF item in the TOC
proc shield/config/get_previous_url { { setting {} } } {
    global sh_config_url_array 
    return $sh_config_url_array([shield/config/get_previous $setting])
}

# determines the next item in the TOC
proc shield/config/get_next { { setting {} } } {
    global sh_config_toc_list

    set this_setting_index [lsearch -exact $sh_config_toc_list $setting]
    if {$this_setting_index == 0} {
        set next_part [lindex $sh_config_toc_list 1]
    } elseif { $this_setting_index == [llength $sh_config_toc_list] } {
        set next_part {}
    } else {
        set next_part [lindex $sh_config_toc_list [expr $this_setting_index + 1]]
    }
    return $next_part
}

# determines the previous item in the TOC
proc shield/config/get_previous { { setting {} } } {
    global sh_config_toc_list
    set this_setting_index [lsearch -exact $sh_config_toc_list $setting]
    
    if {$setting == ""} {
        set next_part [lindex $sh_config_toc_list [expr [llength $sh_config_toc_list] - 1]]
    } elseif {$this_setting_index == 0} {
        set next_part [lindex $sh_config_toc_list [llength $sh_config_toc_list]]
    } else {
        set next_part [lindex $sh_config_toc_list [expr $this_setting_index - 1]]
    }
    return $next_part
}

# for a given TOC item, returns the HREF to go right to it
proc shield/config/get_toc_href { setting } {
    global sh_config_url_array
    catch { return $sh_config_url_array($setting) }
}

proc shield/config/get_page_number { { setting {} } } {
    global sh_config_toc_page_numbers
    return $sh_config_toc_page_numbers($setting)
}

proc shield/config/stop_server {} {
    #catch {
    #return $forever
    #}
    global forever
    set forever 2
    return "ok"
    #return [pid]
}

# builds the array of config items, and a list of config items
proc shield/config/build/sh_config_many_arrays { } {
    global sh_config_array sh_keyword_url sh_file_keywords
    global sh_config_cats sh_config_array_keywords sh_config_array_keywords_inverted
    global shield_dir sh_config_cat_by_setting sh_config_url_array_no_plus
    global sh_config_array_filename sh_config_array_desc sh_config_array_cat sh_config_array_form sh_config_array_justname sh_config_array_tag
    
    global sh_config_url_array sh_config_toc_list sh_config_toc_page_numbers sh_config_page_number_total sh_config_inverted_array
    catch { unset sh_config_url_array sh_config_toc_list sh_config_toc_page_numbers sh_config_page_number_total sh_config_inverted_array}

    set page_number 1
    foreach cat [lsort [array names sh_config_cats]] {
        regsub -all -- { } $cat "+" cat_href 
        
        # this variable holds the Category (plus encoded) as the key, and the URL as the value
        set sh_config_url_array($cat_href) "${shield_dir}/config/viewcategory?category=$cat_href"
        
        # this variable holds the Category (not plus encoded) as the key, and the URL as the value
        set sh_config_url_array_no_plus($cat) "${shield_dir}/config/viewcategory?category=$cat_href"
        
        # we are not putting the category in the inverted array, because we don't yet know
        # how to handle a jump GO! button to a category
        # this variable holds an inverted array of categories and settings, with the description as the key, and the setting as value
        #set sh_config_inverted_array($cat) $cat_href

        lappend sh_config_toc_list $cat_href
        
        incr page_number    
        set sh_config_toc_page_numbers($cat_href) $page_number

        set filenames $sh_config_cats($cat)
        foreach filename [lsort $filenames] {
            set justname $sh_config_array_justname($filename)
            set desc $sh_config_array_desc($justname)
            set tag $sh_config_array_tag($justname)
            set sh_config_url_array($justname) "${shield_dir}/config/edit?setting=$justname"
            set sh_config_inverted_array($tag) $justname
            set sh_config_cat_by_setting($justname) $cat
            lappend sh_config_toc_list $justname
            
            incr page_number    
            set sh_config_toc_page_numbers($justname) $page_number
        }
    }

    # special case for main menu
    set sh_config_url_array() "${shield_dir}/admin"
    set sh_config_toc_page_numbers() 1
    
    # save the number of the last page
    set sh_config_page_number_total $page_number
    
    # build the inverted index term list. The array is the list of index terms as the key, and the documents that they appear in the value field
    foreach file [array names sh_config_array_keywords] {
        set these_kw $sh_config_array_keywords($file)
        set all_kw [split $these_kw ","]
        set sh_file_keywords($file) { }
        
        foreach kw $all_kw {
        # now break up $these_kw and save the results back into an array
            #set kw2 [string trim $kw]
            # global array of each index term, and what file it exists in
            lappend sh_config_array_keywords_inverted([string trim $kw]) [list $file]
            
            # global array of each file and the keywords it has
            lappend sh_file_keywords($file) [string trim $kw]
            #lappend sh_file_keywords($file) [list [string trim $kw]]
        }
    }

    # save the URL for each index keyword. The kw is the key, and the url is the value
    foreach kw [array names sh_config_array_keywords_inverted] {
        regsub -all -- { } $kw "+" kw_href 
        set sh_keyword_url($kw) "${shield_dir}/admin/index/word?word=$kw_href"
    }

    

    return
}

# return the cache copy of the GO button
proc shield/config/get_go_button_cached { } {
    global sh_go_button_cached
    return $sh_go_button_cached
}

# build a cached copy of the GO button
proc shield/config/build/sh_go_button_cached { } {
    global sh_go_button_cached
    catch { unset sh_go_button_cached }
    set sh_go_button_cached [shield/config/get_go_button]
}


# jump to a specific page, based on the value of the GO! button
proc shield/jump { jump } {
    #global sh_config_inverted_array sh_config_array_justname sh_config_array_desc sh_config_array_tag 
    #append retstring $jump 
    #append retstring $sh_config_array_justname($jump) "<br>\n"
    #append retstring $sh_config_array_desc($jump) "<br>\n"
    #append retstring $sh_config_array_tag($jump) "<br>\n"
    #append retstring $sh_config_url_array($jump) "<br>\n"
    
    #return $retstring
    return [shield/config/edit $jump]
}




# returns the table of contents as quick-jump listbox w/a GO button
proc shield/config/get_go_button { } {
    # commented this out until I can figure out how to do server-redirects
    #return;

    global sh_config_inverted_array shield_dir
    
    append retstring "<FORM METHOD=\"GET\" ACTION=\"${shield_dir}/jump\"><SELECT NAME=\"jump\">\n"
    foreach select_option [lsort [array names sh_config_inverted_array]] {
        append retstring "<OPTION value=\"$sh_config_inverted_array($select_option)\">$select_option\n" 
    }
    append retstring "</SELECT> <INPUT TYPE=\"submit\" VALUE=\"Go\"></form>\n"
    return $retstring
}


# parses out the information from a single config file
proc shield/config/parsefile { file } {
    global sh_config
    global sh_config_cats
    global sh_config_array_filename
    global sh_config_array_desc
    global sh_config_array_cat
    global sh_config_array_form
    global sh_config_array_justname
    global sh_config_array_keywords
    global sh_config_array_tag
    global sh_config_array_help
    global sh_config_array_value
    global sh_config_array_instructions
    global sh_config_array_filedate
    
    set justname [shield/config/justname $file]
    set thistext [ReadEntireFile "$sh_config/$file"]
    
    set desc {}
    set cat {}
    set form {}
    set tag {}
    set keywords {}
    set help {}
    set value {}
    set instructions {}
    catch {
        regexp {Description: "([^"]*)"} $thistext retval desc
    }
    catch {
        regexp {Category: "([^"]*)"} $thistext retval cat
    }
    catch {
        regexp {Format: "([^"]*)"} $thistext retval form
    }
    catch {
        regexp {Keywords: "([^"]*)"} $thistext retval keywords
    }
    catch {
        regexp {Instructions: "([^"]*)"} $thistext retval instructions
    }
    catch {
        regexp {\[([^]]*)\]} $thistext retval tag
    }
    catch {
        # extract the help text from this line
        set found_format 0
        set found_instructions 0
        foreach temp_line [split $thistext \n]  {
            #if {[string first "# Instructions: " $temp_line] != -1} {
                # do nothing once we're in the instructions part of the text
            #    set found_format 0
            #}
            
            if {$found_format == 1} {
                if {[string index $temp_line 0] == "#"} {
                    regsub -all -- {#} $temp_line {} temp_line
                    append help "$temp_line !!!"
                } else {
                    append value "$temp_line\n"
                }
            } 
            
            if {[regexp {Format: } $temp_line] == 1} {
                set found_format 1
            }
        }
        regsub -all -- { *!!! *!!! *} $help "<p>\n" help
        regsub -all -- { *!!! *} $help "\n" help
        
        #regsub -all {\#\n} $help {} help
        #regsub -all {:} $help {@} help
    }

    set value [string trimright $value \n]

    set retstring {}
    append retstring "Name: '$file'\n"
    append retstring "Filename: '$justname'\n"
    append retstring "Desc: '$desc'\n"
    append retstring "Tag: '$tag'\n"
    append retstring "Cat: '$cat'\n"
    append retstring "Form: '$form'\n"
    append retstring "--------------------------------------------------------------------\n"
    
    #WriteToLog "Name: '$file'"
    #WriteToLog "Filaname: '$justname'"
    #WriteToLog "Desc: '$desc'"
    #WriteToLog "Cat: '$cat'"
    #WriteToLog "Form: '$form'"
    #WriteToLog "--------------------------------------------------------------------"
    
    if {[string length $desc] == 0} {
        error "Error: no description in file '$file' "
    }
    if {[string length $cat] == 0} {
        error "Error: no category in file '$file' "
    }
    if {[string length $form] == 0} {
        error "Error: no format in file '$file' "
    }
    
    set sh_config_array_justname($file) $justname
    set sh_config_array_filename($justname) $file
    set sh_config_array_desc($justname) $desc
    set sh_config_array_cat($justname) $cat
    set sh_config_array_tag($justname) $tag
    set sh_config_array_form($justname) $form
    set sh_config_array_keywords($justname) $keywords
    set sh_config_array_help($justname) $help
    set sh_config_array_value($justname) $value
    set sh_config_array_instructions($justname) $instructions
    set sh_config_array_filedate($file) [file atime $sh_config/$file]
    
    append sh_config_cats($cat) "$file "
    
    return $retstring
}

# gets just the name of the passed file, stripping off the filename extension
proc shield/config/justname { file } {
    set justname {}
    if [catch {
        regexp {(.*)\..*} $file retval justname
    }] {
        WriteToLog "! justname failed on $file"
    }
    return $justname
}

# reload the web interface with the new port # or tcpip address
proc shield/config/reload_server {} {
    global sh_config_array_value Config

    # if the tcpip address is set to 0 then disable the web interface
    if {$sh_config_array_value(webtcpip) != 0} {
    
        # if the port number not set then set it at 81 by default
        if {$sh_config_array_value(webport) == ""} {
            set sh_config_array_value(webport) 81
        }
    
        Httpd_Shutdown
        Httpd_Server $sh_config_array_value(webport) $Config(host) $sh_config_array_value(webtcpip)
        WriteToLog \n\[[clock format [clock seconds]]\]
        WriteToLog "Restarting web server on port $sh_config_array_value(webport)"
    } else {
        Httpd_Shutdown
    }
}

proc shield/config/save { setting parameters } {
    global sh_config_array_value sh_config sh_config_array_filename
    global sh_full_text_body
    
    # get all the commented line (with an #) in the variable new_line
    set bool 1
    set text [ReadEntireFile $sh_config/$sh_config_array_filename($setting)]
    foreach line [split $text \n] {
        if {$bool == 1} {
            append new_text $line
            append new_text "\n"
        }
        if {[string first "# Instructions:" $line] == 0} {
            set bool 0
        }
    }

    # I have to suppress all the carriage return because IE5.0 put a carriage
    # return after all lines of a textarea. Tcl interpret them as a new line!
    # regsub -all \x0d $parameters {} parameters
    regsub -all \x0d $parameters {} parameters
    regsub -all \x0d $new_text {} new_text
    
    append new_text $parameters
    
    # set the new parameters value in the server memory
    set sh_config_array_value($setting) $parameters

    # update the sh_full_text_body($setting) for the search function
    set sh_full_text_body($setting) $new_text
    
    # save the parameters in a new file
    SaveEntireFile $sh_config/$sh_config_array_filename($setting) $new_text

    # reload the web server if the port # or the tcpip address changed
    if {($setting == "webport") || ($setting == "webtcpip") || ($setting == "template")} {
        shield/config/reload_server
    }

    #return "<html><body><pre>$new_text</pre></body></html>"
    shield/config/edit $setting
}


# lists all the files in the "rules" directory
proc shield/rules/listfiles { } {
    global sh_rules
    return [GetFilesForDir $sh_rules]
}

# lists all the files in the "config" directory
proc shield/config/listfiles { } {
    global sh_config
    return [GetFilesForDir $sh_config]
}

# given a directory name, returns a list of files in that directory
proc GetFilesForDir { dirname } {
    set old_dir [pwd]
    cd $dirname
    set retval [glob -nocomplain *]
    cd $old_dir
    return $retval
}

# read the entire contents of a file into memory and returns it
proc ReadEntireFile { filename } {
    #WriteToLog "reading file: $filename"
    set file_text ""
    catch {
        set file_id [open $filename RDONLY]
        set file_text [read $file_id]
        close $file_id
    }
    return $file_text
}

proc SaveEntireFile { filename text } {
    global sh_config sh_config_array_filedate
    catch {
        set file_id [open $filename w]
        puts $file_id $text
        close $file_id
    }
    set sh_config_array_filedate([file tail $filename]) [file atime $filename]
    return 1
}

proc WriteToLog { str } {
    if {$str != ""} {
        global web_dir

        set file_id {}
        if {[file exists $web_dir/log.txt]} {
            set file_id [open $web_dir/log.txt a]
        } else {
            set file_id [open $web_dir/log.txt w]
        }

        puts $file_id $str
        close $file_id
    }
}


